@@ -8,17 +8,18 @@
 # copied, modified, or distributed except according to those terms.
 """Python Library to manage NetHSM(s)."""
 
-__version__ = "1.0.0"
+__version__ = "1.1.0"
 
 import binascii
 import contextlib
 import enum
 import json
+import string
 from base64 import b64decode, b64encode
 from dataclasses import dataclass
 from datetime import datetime, timezone
 from io import BufferedReader, FileIO
-from typing import TYPE_CHECKING, Any, Iterator, Mapping, Optional, Union
+from typing import TYPE_CHECKING, Any, Iterator, Mapping, NoReturn, Optional, Union
 from urllib.parse import urlencode
 
 import urllib3
@@ -205,14 +206,43 @@ class Base64:
         return b64decode(self.data)
 
     @classmethod
-    def from_encoded(cls, data: Union[bytes, str]) -> "Base64":
+    def from_encoded(
+        cls, data: Union[bytes, str], ignore_whitespace: bool = False
+    ) -> "Base64":
+        """
+        >>> Base64.from_encoded("dGVzdAo=")
+        Base64(data='dGVzdAo=')
+        >>> Base64.from_encoded(b"dGVzdAo=")
+        Base64(data='dGVzdAo=')
+        >>> Base64.from_encoded("dGV zdAo=")
+        Traceback (most recent call last):
+            ...
+        ValueError: Invalid base64 data: Non-base64 digit found: dGV zdAo=
+        >>> Base64.from_encoded(b"dGV zdAo=")
+        Traceback (most recent call last):
+            ...
+        ValueError: Invalid base64 data: Non-base64 digit found: dGV zdAo=
+        >>> Base64.from_encoded("dGV zdAo=", ignore_whitespace=True)
+        Base64(data='dGVzdAo=')
+        >>> Base64.from_encoded(b"dGV zdAo=", ignore_whitespace=True)
+        Base64(data='dGVzdAo=')
+        """
+        if ignore_whitespace:
+            if isinstance(data, bytes):
+                data = data.translate(None, delete=string.whitespace.encode())
+            else:
+                data = data.translate(str.maketrans("", "", string.whitespace))
+
         try:
             b64decode(data, validate=True)
+        except binascii.Error as e:
             if isinstance(data, bytes):
-                data = data.decode()
-            return cls(data=data)
-        except binascii.Error:
-            raise ValueError(f"Invalid base64 data: {data!r}")
+                data = data.decode(errors="replace")
+            raise ValueError(f"Invalid base64 data: {e}: {data}") from None
+
+        if isinstance(data, bytes):
+            data = data.decode()
+        return cls(data=data)
 
     @classmethod
     def encode(cls, data: bytes) -> "Base64":
@@ -310,7 +340,7 @@ def _handle_exception(
     messages: dict[int, str] = {},
     roles: list[Role] = [],
     state: Optional[State] = None,
-) -> None:
+) -> NoReturn:
     from .client import ApiException
 
     if isinstance(e, ApiException):
@@ -328,7 +358,7 @@ def _handle_api_exception(
     messages: dict[int, str] = {},
     roles: list[Role] = [],
     state: Optional[State] = None,
-) -> None:
+) -> NoReturn:
     if e.status in messages:
         message = messages[e.status]
         raise NetHSMError(message)
@@ -631,8 +661,6 @@ class NetHSM:
                     409: f"Conflict -- a user with the ID {user_id} already exists",
                 },
             )
-        if user_id is None:
-            raise NetHSMError("Could not determine the ID of the new user")
         return user_id
 
     def delete_user(self, user_id: str) -> None:
@@ -981,7 +1009,69 @@ class NetHSM:
                     409: f"Conflict -- a key with the ID {key_id} already exists",
                 },
             )
-        assert key_id
+        return key_id
+
+    def add_key_pem(
+        self,
+        key_id: Optional[str],
+        mechanisms: list[KeyMechanism],
+        private_key: str,
+        tags: list[str] = [],
+    ) -> str:
+        from .client.components.schema.key_mechanisms import KeyMechanismsTupleInput
+        from .client.components.schema.key_restrictions import KeyRestrictionsDict
+        from .client.components.schema.private_key_pem import (
+            ArgumentsDict,
+            PrivateKeyPemDict,
+        )
+        from .client.components.schema.tag_list import TagListTuple
+
+        mechanism_tuple: KeyMechanismsTupleInput = [
+            mechanism.value for mechanism in mechanisms
+        ]
+
+        if tags:
+            arguments = ArgumentsDict(
+                mechanisms=mechanism_tuple,
+                restrictions=KeyRestrictionsDict(
+                    tags=TagListTuple([tag for tag in tags])
+                ),
+            )
+        else:
+            arguments = ArgumentsDict(
+                mechanisms=mechanism_tuple,
+            )
+        body = PrivateKeyPemDict(arguments=arguments, key_file=private_key)
+
+        try:
+            if key_id:
+                from .client.paths.keys_key_id.put.path_parameters import (
+                    PathParametersDict,
+                )
+
+                path_params = PathParametersDict(KeyID=key_id)
+                self._get_api().keys_key_id_put(
+                    path_params=path_params,
+                    body=body,
+                    content_type="multipart/form-data",
+                )
+            else:
+                response = self._get_api().keys_post(
+                    body=body,
+                    content_type="multipart/form-data",
+                    skip_deserialization=True,
+                )
+                key_id = self._get_create_resource_id(response.response)
+        except Exception as e:
+            _handle_exception(
+                e,
+                state=State.OPERATIONAL,
+                roles=[Role.ADMINISTRATOR],
+                messages={
+                    400: "Bad request -- specified properties are invalid",
+                    409: f"Conflict -- a key with the ID {key_id} already exists",
+                },
+            )
         return key_id
 
     def delete_key(self, key_id: str) -> None:
